import networkx as nx
from tuw_nlp.graph.utils import graph_to_pn
from xpotato.dataset.utils import default_pn_to_graph


class PotatoGraph:
    def __init__(self, graph: nx.DiGraph = None, graph_str: str = None) -> None:
        if graph:
            self.graph = graph
        elif graph_str:
            self.graph, _ = default_pn_to_graph(graph_str)
        else:
            self.graph = None

    def __str__(self) -> str:
        return graph_to_pn(self.graph)



from typing import Dict, List, Tuple

import pandas as pd
from xpotato.dataset.dataset import Dataset
from xpotato.dataset.relation_sample import RelationSample


class RelationDataset(Dataset):
    def __init__(
        self, examples: List[Tuple[str, str]], label_vocab: Dict[str, int], lang="en"
    ) -> None:
        super().__init__(examples=examples, label_vocab=label_vocab, lang=lang)

    def read_dataset(self, examples: List[Tuple[str, str]]) -> List[RelationSample]:
        return [RelationSample(example) for example in examples]

    def to_dataframe(self) -> pd.DataFrame:
        df = pd.DataFrame(
            {
                "text": [sample.text for sample in self._dataset],
                "label": [sample.label for sample in self._dataset],
                "label_id": [
                    self.label_vocab[sample.label] if sample.label else None
                    for sample in self._dataset
                ],
                "entity1": [sample.entity1 for sample in self._dataset],
                "entity2": [sample.entity2 for sample in self._dataset],
                "graph": [sample.graph for sample in self._dataset],
            }
        )
        return df



import json
import os
from collections import defaultdict

import networkx as nx
import pandas as pd
import stanza
import penman as pn
from networkx.algorithms.isomorphism import DiGraphMatcher
from sklearn.metrics import precision_recall_fscore_support
from tqdm import tqdm
from tuw_nlp.grammar.text_to_4lang import TextTo4lang
from tuw_nlp.graph.utils import (
    GraphFormulaMatcher,
)
from tuw_nlp.text.pipeline import CachedStanzaPipeline

from xpotato.dataset.utils import default_pn_to_graph, ud_to_graph, amr_pn_to_graph


class GraphExtractor:
    def __init__(self, cache_dir=None, cache_fn=None, lang=None):
        self.cache_dir = cache_dir
        self.cache_fn = cache_fn
        self.lang = lang
        self.nlp = None
        self.matcher = None
        self.amr_stog = None

    def init_amr(self):
        if self.amr_stog == None:
            import amrlib

            self.amr_stog = amrlib.load_stog_model()

    def init_nlp(self):
        if self.nlp == None:
            if self.lang == "en_bio":
                nlp = stanza.Pipeline("en", package="craft")
            else:
                nlp = stanza.Pipeline(self.lang)
            self.nlp = CachedStanzaPipeline(nlp, self.cache_fn)

    def parse_iterable(self, iterable, graph_type="fourlang", lang=None):
        if lang:
            self.lang = lang
        if graph_type == "fourlang":
            with TextTo4lang(
                lang=self.lang, nlp_cache=self.cache_fn, cache_dir=self.cache_dir
            ) as tfl:
                for sen in tqdm(iterable):
                    fl_graphs = list(tfl(sen))
                    g = fl_graphs[0]
                    for n in fl_graphs[1:]:
                        g = nx.compose(g, n)
                    yield g

        elif graph_type == "ud":
            self.init_nlp()
            for sen in tqdm(iterable):
                doc = self.nlp(sen)
                g, _ = ud_to_graph(doc.sentences[0])
                for doc_sen in doc.sentences[1:]:
                    n, _ = ud_to_graph(doc_sen)
                    g = nx.compose(g, n)
                yield g

        elif graph_type == "amr":
            self.init_amr()
            for sen in tqdm(iterable):
                graphs = self.amr_stog.parse_sents([sen])
                g, _ = amr_pn_to_graph(graphs[0])
                yield g


class FeatureEvaluator:
    def __init__(self, graph_format="ud"):
        self.graph_format = graph_format

    # ADAM: Very important to assign IDs to features from 0 because that's how
    # the mapping will work!!
    def annotate(self, graph, features):
        feature_to_marked_nodes = {}

        for i, feature in enumerate(features):
            assert (
                len(feature) == 4
            ), f"Feature must be a 4-tuple for OpenIE, not {feature}"

            positive_features = feature[0]
            negative_features = feature[1]

            for positive in positive_features:
                p = pn.decode(positive)
                first = p.triples[0][0]
                assert first == "u_0", f"The IDs must start from 0, not {first}"

            for negative in negative_features:
                p = pn.decode(negative)
                first = p.triples[0][0]
                assert first == "u_0", f"The IDs must start from 0, not {first}"

            feature_to_marked_nodes[i] = feature[3]
            features[i] = feature[:3]

        matcher = GraphFormulaMatcher(features, converter=default_pn_to_graph)
        feats = matcher.match(graph, return_subgraphs=True)

        for key, i, subgraphs in feats:
            triplet = {"relation": key}
            marked_nodes = feature_to_marked_nodes[i]
            for j, node in enumerate(marked_nodes):
                subgraph = subgraphs[j]

                node_to_node = {}
                for id, graph_node in subgraph.nodes(data=True):
                    mapping = graph_node["mapping"]
                    node_to_node[mapping] = graph_node["name"]

                for k, v in node.items():
                    triplet[k] = node_to_node[v]

                yield triplet

    def annotate_dataframe(self, dataset, features):
        graphs = dataset.graph.tolist()

        triplets = []
        for graph in graphs:
            relations = self.annotate(graph, features)
            triplets.append(list(relations))
        d = {
            "Sentence": dataset.text.tolist(),
            "Triplets": triplets,
        }

        return pd.DataFrame(d)

    def match_features(self, dataset, features, multi=False, return_subgraphs=False):
        graphs = dataset.graph.tolist()

        matches = []
        predicted = []
        matched_graphs = []

        matcher = GraphFormulaMatcher(features, converter=default_pn_to_graph)

        for i, g in tqdm(enumerate(graphs)):
            feats = matcher.match(g, return_subgraphs=return_subgraphs)
            if multi:
                self.match_multi(feats, features, matches, predicted, matched_graphs)
            else:
                self.match_not_multi(
                    feats, features, matches, predicted, matched_graphs
                )

        d = {
            "Sentence": dataset.text.tolist(),
            "Predicted label": predicted,
            "Matched rule": matches,
        }
        if return_subgraphs:
            d["Matched subgraph"] = matched_graphs

        df = pd.DataFrame(d)
        return df

    def match_multi(self, feats, features, matches, predicted, matched_graphs):
        keys = []
        matched_rules = []
        matched_subgraphs = []
        for key, feature, graphs in feats:
            if key not in keys:
                matched_rules.append(features[feature])
                matched_subgraphs.append(graphs)
                keys.append(key)
        if not keys:
            matches.append("")
            predicted.append("")
            graphs.append("")
        else:
            matches.append(matched_rules)
            predicted.append(keys)
            matched_graphs.append(matched_subgraphs)

    def match_not_multi(self, feats, features, matches, predicted, matched_graphs):
        for key, feature, graphs in feats:
            matches.append(features[feature])
            predicted.append(key)
            matched_graphs.append(graphs)
            break
        else:
            matches.append("")
            matched_graphs.append("")
            predicted.append("")

    def one_versus_rest(self, df, entity):
        mapper = {entity: 1}

        one_versus_rest_df = df.copy()
        one_versus_rest_df["one_versus_rest"] = [
            mapper[item] if item in mapper else 0 for item in df.label
        ]

        return one_versus_rest_df

    def rank_features(self, cl, features, orig_data, false_negatives):
        # TODO: currently disabled
        # if false_negatives:
        #     subset_data = orig_data.iloc[false_negatives]
        # else:
        #   subset_data = orig_data
        subset_data = orig_data
        df, accuracy = self.evaluate_feature(cl, features, subset_data)

        features_stat = []

        for i, feature in enumerate(features):
            features_stat.append(
                (
                    feature,
                    df.iloc[i].Precision,
                    df.iloc[i].Recall,
                    df.iloc[i].Fscore,
                    df.iloc[i].Support,
                    len(df.iloc[i].True_positive_sens),
                    len(df.iloc[i].False_positive_sens),
                )
            )

        def rank(feature):
            return len(df.iloc[features.index(feature[0])].True_positive_graphs)

        return sorted(features_stat, key=rank, reverse=True)

    def train_feature(self, cl, feature, data, graph_format="ud"):
        feature_graph = default_pn_to_graph(feature)[0]

        graphs = data.graph.tolist()
        labels = self.one_versus_rest(data, cl).one_versus_rest.tolist()
        path = "trained_features.tsv"
        trained_features = []
        with open(path, "w+") as f:
            for i, g in enumerate(graphs):
                matcher = DiGraphMatcher(
                    g,
                    feature_graph,
                    node_match=GraphFormulaMatcher.node_matcher,
                    edge_match=GraphFormulaMatcher.edge_matcher,
                )
                if matcher.subgraph_is_monomorphic():
                    for iso_pairs in matcher.subgraph_monomorphisms_iter():
                        nodes = []
                        for k in iso_pairs:
                            if not nodes:
                                if feature_graph.nodes[iso_pairs[k]]["name"] == ".*":
                                    nodes.append(g.nodes[k]["name"])
                        if not nodes:
                            g2_to_g1 = {v: u for (u, v) in iso_pairs.items()}
                            for u, v, attrs in feature_graph.edges(data=True):
                                if attrs["color"] == ".*":
                                    edge = g.get_edge_data(g2_to_g1[u], g2_to_g1[v])[
                                        "color"
                                    ]
                                    nodes.append(edge)
                        nodes_str = ",".join(nodes)
                        label = labels[i]
                        sentence = data.iloc[i].text
                        f.write(f"{feature}\t{nodes_str}\t{sentence}\t{label}\n")
                        trained_features.append(
                            (feature, nodes_str, sentence, str(label))
                        )

        return self.cluster_feature(trained_features)

    def cluster_feature(self, trained_features):
        graphs = {}
        if os.path.isfile("longman_zero_paths_one_exp"):
            with open("longman_zero_paths_one_exp.json") as f:
                graphs = json.load(f)

        words = {}
        for fields in trained_features:
            words[fields[1] + "_" + fields[3]] = int(fields[3])
            feature = fields[0]
        graph = nx.MultiDiGraph()

        for word in words:
            if words[word] == 1:
                color = "green"
            else:
                color = "red"
            graph.add_node(word, color=color)
            word_clean = word.split("_")[0]
            if word_clean in graphs:
                hypernyms = graphs[word_clean]
                for hypernym in hypernyms:
                    hypernym_words = hypernyms[hypernym]
                    for w in hypernym_words:
                        if hypernym == "1":
                            graph.add_edge(word, w, color=hypernym)

        selected_words = self.select_words(trained_features)

        word_features = []

        if selected_words:
            word_features.append(feature.replace(".*", "|".join(selected_words), 1))
        else:
            word_features.append(feature)

        return word_features

    def select_words(self, trained_features):
        features = []
        labels = []

        for fields in trained_features:
            features.append(fields[1])
            labels.append(int(fields[3]))
        words_to_measures = {
            word: {"TP": 0, "FP": 0, "TN": 0, "FN": 0} for word in set(features)
        }
        for word in words_to_measures:
            for i, label in enumerate(labels):
                if label and features[i] == word:
                    words_to_measures[word]["TP"] += 1
                if label and features[i] != word:
                    words_to_measures[word]["FN"] += 1
                if not label and features[i] == word:
                    words_to_measures[word]["FP"] += 1
                if not label and features[i] != word:
                    words_to_measures[word]["TN"] += 1

        for word in words_to_measures:
            TP = words_to_measures[word]["TP"]
            FP = words_to_measures[word]["FP"]
            TN = words_to_measures[word]["TN"]
            FN = words_to_measures[word]["FN"]

            precision = TP / (TP + FP)
            recall = TP / (TP + FN)

            words_to_measures[word]["precision"] = precision
            words_to_measures[word]["recall"] = recall

        selected_words = set()

        for word in words_to_measures:
            if words_to_measures[word]["precision"] > 0.9 and (
                words_to_measures[word]["TP"] > 1
                or words_to_measures[word]["recall"] > 0.01
            ):
                selected_words.add(word)

        return selected_words

    def evaluate_feature(self, cl, features, data, graph_format="ud"):
        measure_features = []
        graphs = data.graph.tolist()
        labels = self.one_versus_rest(data, cl).one_versus_rest.tolist()

        whole_predicted = []
        matched = defaultdict(list)

        # We want to view false negative examples for all rules, not rule specific
        false_neg_g = []
        false_neg_s = []
        false_neg_indices = []
        matcher = GraphFormulaMatcher(features, converter=default_pn_to_graph)
        for i, g in enumerate(graphs):
            feats = matcher.match(g)
            label = 0
            for key, feature in feats:
                matched[i].append(features[feature][0])
                label = 1
            whole_predicted.append(label)

            if label == 0 and labels[i] == 1:
                false_neg_g.append(g)
                sen = data.iloc[i].text
                lab = data.iloc[i].label
                false_neg_s.append((sen, lab))
                false_neg_indices.append(i)

        accuracy = []
        for pcf in precision_recall_fscore_support(
            labels, whole_predicted, average=None
        ):
            if len(pcf) > 1:
                accuracy.append(pcf[1])
            else:
                accuracy.append(0)

        for feat in features:
            measure = [feat[0]]
            false_pos_g = []
            false_pos_s = []
            false_pos_indices = []
            true_pos_g = []
            true_pos_s = []
            true_pos_indices = []
            predicted = []
            for i, g in enumerate(graphs):
                feats = matched[i]
                label = 1 if feat[0] in feats else 0
                if label == 1 and labels[i] == 0:
                    false_pos_g.append(g)
                    sen = data.iloc[i].text
                    lab = data.iloc[i].label
                    false_pos_s.append((sen, lab))
                    false_pos_indices.append(i)
                if label == 1 and labels[i] == 1:
                    true_pos_g.append(g)
                    sen = data.iloc[i].text
                    lab = data.iloc[i].label
                    true_pos_s.append((sen, lab))
                    true_pos_indices.append(i)
                predicted.append(label)
            for pcf in precision_recall_fscore_support(labels, predicted, average=None):
                if len(pcf) > 1:
                    measure.append(pcf[1])
                else:
                    measure.append(0)
            measure.append(false_pos_g)
            measure.append(false_pos_s)
            measure.append(false_pos_indices)
            measure.append(true_pos_g)
            measure.append(true_pos_s)
            measure.append(true_pos_indices)
            measure.append(false_neg_g)
            measure.append(false_neg_s)
            measure.append(false_neg_indices)
            measure.append(predicted)
            measure_features.append(measure)

        df = pd.DataFrame(
            measure_features,
            columns=[
                "Feature",
                "Precision",
                "Recall",
                "Fscore",
                "Support",
                "False_positive_graphs",
                "False_positive_sens",
                "False_positive_indices",
                "True_positive_graphs",
                "True_positive_sens",
                "True_positive_indices",
                "False_negative_graphs",
                "False_negative_sens",
                "False_negative_indices",
                "Predicted",
            ],
        )

        return df, accuracy



from collections import defaultdict

import networkx as nx
import pandas as pd
import penman as pn
from tuw_nlp.graph.utils import preprocess_node_alto
from tuw_nlp.graph.utils import graph_to_pn


def save_dataframe(df: pd.DataFrame, path: str) -> None:
    df_to_save = df.copy()
    graphs = [graph_to_pn(graph) for graph in df_to_save["graph"].tolist()]
    df_to_save["graph"] = graphs
    df_to_save.to_csv(path, index=False, sep="\t")


def ud_to_graph(sen, edge_attr="color"):
    """convert dependency-parsed stanza Sentence to nx.DiGraph"""
    G = nx.DiGraph()
    root_id = None
    for word in sen.to_dict():
        if isinstance(word["id"], (list, tuple)):
            # token representing an mwe, e.g. "vom" ~ "von dem"
            continue
        G.add_node(word["id"], name=preprocess_node_alto(word["lemma"]))
        if word["deprel"] == "root":
            root_id = word["id"]
            G.add_node(word["head"], name="root")
        G.add_edge(word["head"], word["id"])
        G[word["head"]][word["id"]].update(
            {edge_attr: preprocess_node_alto(word["deprel"])}
        )

    return G, root_id


def default_pn_to_graph(raw_dl, edge_attr="color"):
    g = pn.decode(raw_dl)
    G = nx.DiGraph()

    char_to_id = defaultdict(int)
    next_id = 0
    for i, trip in enumerate(g.triples):
        if i == 0:
            root_id = next_id
            name = trip[2]
            if not name:
                name = "None"
            G.add_node(root_id, name=name)
            char_to_id[trip[0]] = next_id
            next_id += 1

        elif trip[1] == ":instance":
            if trip[2]:
                name = trip[2]
                if not name:
                    name = "None"
                G.add_node(next_id, name=name)
                char_to_id[trip[0]] = next_id
                next_id += 1

    for trip in g.triples:
        if trip[1] != ":instance":
            edge = trip[1].split(":")[1]
            src = trip[0]
            tgt = trip[2]
            if src not in char_to_id:
                char_to_id[src] = next_id
                name = src
                if not name:
                    name = "None"
                G.add_node(next_id, name=name)
                next_id += 1
            if tgt not in char_to_id:
                char_to_id[tgt] = next_id
                name = tgt
                if not name:
                    name = "None"
                G.add_node(next_id, name=name)
                next_id += 1

            G.add_edge(char_to_id[src], char_to_id[tgt])
            G[char_to_id[src]][char_to_id[tgt]].update({edge_attr: edge})

    return G, root_id


def amr_pn_to_graph(raw_dl, edge_attr="color", clean_nodes=False):
    g = pn.decode(raw_dl)
    G = nx.DiGraph()

    char_to_id = defaultdict(int)
    next_id = 0
    for i, trip in enumerate(g.triples):
        if i == 0:
            root_id = next_id
            if "-" in trip[2]:
                name = "-".join(trip[2].split("-")[:-1])
            else:
                name = trip[2]
            if clean_nodes:
                if name:
                    name = preprocess_node_alto(name)
                else:
                    name = "None"
            G.add_node(root_id, name=name)
            char_to_id[trip[0]] = next_id
            next_id += 1

        elif trip[1] == ":instance":
            if trip[2]:
                if "-" in trip[2]:
                    name = "-".join(trip[2].split("-")[:-1])
                else:
                    name = trip[2]
                if clean_nodes:
                    if name:
                        name = preprocess_node_alto(name)
                    else:
                        name = "None"
                G.add_node(next_id, name=name)
                char_to_id[trip[0]] = next_id
                next_id += 1

    for trip in g.triples:
        if trip[1] != ":instance":
            edge = "".join(trip[1].split(":")[1].split("-"))
            src = trip[0]
            tgt = trip[2]
            if src not in char_to_id:
                char_to_id[src] = next_id
                if "-" in src:
                    name = "-".join(src.split("-")[:-1])
                else:
                    name = src
                if clean_nodes:
                    if name:
                        name = preprocess_node_alto(name)
                    else:
                        name = "None"
                G.add_node(next_id, name=name)
                next_id += 1
            if tgt not in char_to_id:
                char_to_id[tgt] = next_id
                if "-" in tgt:
                    name = "-".join(tgt.split("-")[:-1])
                else:
                    name = tgt
                if clean_nodes:
                    if name:
                        name = preprocess_node_alto(name)
                    else:
                        name = "None"
                G.add_node(next_id, name=name)
                next_id += 1

            G.add_edge(char_to_id[src], char_to_id[tgt])
            G[char_to_id[src]][char_to_id[tgt]].update({edge_attr: edge})

    return G, root_id



from collections import defaultdict
from typing import Dict, List
from typing import Union
from math import log2, sqrt
from rank_bm25 import BM25Okapi

import eli5
import pandas as pd
from xpotato.graph_extractor.extract import GraphExtractor, FeatureEvaluator
from xpotato.models.model import GraphModel
from sklearn.linear_model import LogisticRegression
from tqdm import tqdm

import skcriteria as skc
from skcriteria.madm import simple
from skcriteria.pipeline import mkpipe
from skcriteria.preprocessing import invert_objectives, scalers


class GraphTrainer:
    def __init__(
        self,
        dataset: pd.DataFrame,
        lang: str = "en",
        max_edge: int = 2,
        max_features: int = 2000,
    ) -> None:
        print("Initializing trainer object...")
        self.dataset = dataset
        self.extractor = GraphExtractor(lang=lang, cache_fn="en_nlp_cache")
        self.evaluator = FeatureEvaluator()
        self.graph_model = GraphModel()
        self.max_edge = max_edge
        self.max_features = max_features
        self.model = LogisticRegression(random_state=0)

    def get_n_most_similar(
        self,
        sample: str,
        dataset: pd.DataFrame = None,
        n: int = 10,
        algorithm: str = "bm25",
    ) -> List[str]:
        corpus = self.dataset.text.tolist() if not dataset else dataset
        if algorithm == "bm25":
            tokenized_corpus = [doc.split() for doc in corpus]
            bm25 = BM25Okapi(tokenized_corpus)

            query = sample.split()
            top_n = bm25.get_top_n(query, corpus, n=n)

            return top_n

        return corpus[0:10]

    def prepare_and_train(
        self, min_edge: int = 0, rank: bool = False
    ) -> Dict[str, List[List[Union[List[str], str]]]]:
        self.prepare()
        return self.train(min_edge=min_edge, rank=rank)

    def prepare(self) -> None:
        ids = pd.to_numeric(self.dataset.index).tolist()
        sentences = self.dataset.text.tolist()
        labels = self.dataset.label_id.tolist()
        graphs = self.dataset.graph.tolist()

        print(f"Featurizing graphs by generating subgraphs up to {self.max_edge}...")
        for ind, graph, label in tqdm(zip(ids, graphs, labels)):
            self.graph_model.featurize_sen_graph(ind, graph, label, self.max_edge)

        print("Getting feature graphs...")
        self.feature_graphs = self.graph_model.get_feature_graphs()
        self.feature_graph_strings = self.graph_model.get_feature_graph_strings()

        print("Selecting the best features...")
        if self.max_features > len(graphs):
            n_best = int(log2(len(graphs)) * sqrt(len(graphs)))
            self.graph_model.select_n_best(n_best)
        else:
            self.graph_model.select_n_best(self.max_features)

    def rank(
        self, features: Dict[str, List[List[Union[List[str], str]]]]
    ) -> Dict[str, List[List[Union[List[str], str]]]]:

        ranked_features = {}

        for key, val in features.items():
            stat, acc = self.evaluator.evaluate_feature(key, val, self.dataset)
            stat_opt = pd.DataFrame(
                {
                    "false_positives": stat.False_positive_sens.apply(lambda x: len(x)),
                    "true_positives": stat.True_positive_sens.apply(lambda x: len(x)),
                }
            )

            criteria_data = skc.mkdm(
                matrix=stat_opt,
                objectives=[min, max],
                criteria=stat_opt.columns,
                weights=[30, 70],
            )

            dm = mkpipe(
                invert_objectives.MinimizeToMaximize(),
                scalers.SumScaler(target="both"),
                simple.WeightedSumModel(),
            )

            dec = dm.evaluate(criteria_data)

            sorted_feats = [
                x for _, x in sorted(zip(dec.rank_, val), key=lambda pair: pair[0])
            ]
            ranked_features[key] = sorted_feats

        return ranked_features

    def train(
        self, min_edge: int = 1, rank: bool = False
    ) -> Dict[str, List[List[Union[List[str], str]]]]:
        label_vocab = {}
        for label in self.dataset.label.unique():
            label_vocab[label] = (
                self.dataset[self.dataset.label == label].iloc[0].label_id
            )

        inv_vocab = {v: k for k, v in label_vocab.items()}

        print("Generating training data...")
        train_X, train_Y = self.graph_model.get_x_y(
            self.dataset.label.tolist(), label_vocab=label_vocab
        )

        print("Training...")
        self.model.fit(train_X, train_Y)
        weights_df = eli5.explain_weights_df(self.model)
        features = defaultdict(list)

        print("Getting features...")
        for target in weights_df.target.unique():
            targeted_df = weights_df[weights_df.target == target]
            most_important_weights = []

            for i, w in enumerate(targeted_df.weight.tolist()):
                if w > 0.01:
                    most_important_weights.append(
                        targeted_df.iloc[i].feature.strip("x")
                    )

            for i in most_important_weights:
                if i != "<BIAS>":
                    g_nx = self.feature_graphs[self.graph_model.inverse_relabel[int(i)]]
                    g = self.feature_graph_strings[
                        self.graph_model.inverse_relabel[int(i)]
                    ]
                    if len(g_nx.edges()) >= min_edge:
                        features[inv_vocab[int(target)]].append(
                            ([g], [], inv_vocab[int(target)])
                        )

        if rank:
            print("Ranking features based on accuracy...")
            features = self.rank(features)

        return features



from xpotato.dataset.utils import default_pn_to_graph
from xpotato.graph_extractor.extract import FeatureEvaluator
from xpotato.graph_extractor.rule import RuleSet, Rule

import os

dir_name = os.path.dirname(os.path.realpath(__file__))

FEATURE1 = [["(u_0 / find :obl (u_1 / .*) :obj (u_2 / .*))"], [], "FIND"]
FEATURE2 = [["(u_0 / write :nsubj (u_1 / .*) :obj (u_2 / .*))"], [], "WRITE"]

FEATURE3 = [
    ["(u_0 / find :obl (u_1 / .*) :obj (u_2 / .*))"],
    [],
    "FIND",
    [{"ARG1": 1, "ARG2": 2}],
]
FEATURE4 = [
    ["(u_0 / write :nsubj (u_1 / .*) :obj (u_2 / .*))"],
    [],
    "WRITE",
    [{"ARG1": 1, "ARG2": 2}],
]

FEATURE_DICT = {
    "FIND": [[["(u_0 / find :obl (u_1 / .*) :obj (u_2 / .*))"], [], "FIND"]],
    "WRITE": [[["(u_0 / write :nsubj (u_1 / .*) :obj (u_2 / .*))"], [], "WRITE"]],
}

GRAPH = "(u_2 / write  :nsubj (u_1 / person)  :obj (u_4 / sentence  :det (u_3 / this))  :conj (u_6 / find  :cc (u_5 / and)  :obj (u_7 / object)  :obl (u_9 / location  :case (u_8 / in)))  :punct (u_10 / PERIOD)  :root-of (u_0 / root))"


def test_rule():
    rule1 = Rule(FEATURE1)

    assert rule1.to_list() == FEATURE1


def test_ruleset_to_list():
    rule_set = RuleSet([Rule(FEATURE1)])

    rule_set.add_rule(Rule(FEATURE2))

    assert rule_set.to_list() == [FEATURE1, FEATURE2]


def test_ruleset_to_dict():
    rule_set = RuleSet([Rule(FEATURE1), Rule(FEATURE2)])

    assert rule_set.to_dict() == {"FIND": [FEATURE1], "WRITE": [FEATURE2]}


def test_ruleset_from_dict_to_list():
    rule_set = RuleSet()
    rule_set.from_dict(FEATURE_DICT)

    assert rule_set.to_list() == [FEATURE1, FEATURE2]


def test_ruleset_json():
    rule_set = RuleSet()

    rule_set.from_json(os.path.join(dir_name, "features.json"))

    assert rule_set.to_list() == [FEATURE1, FEATURE2]


def test_ruleset_to_tsv():
    rule_set = RuleSet([Rule(FEATURE1), Rule(FEATURE2)])

    rule_set.to_tsv(os.path.join(dir_name, "features.tsv"))

    rule_set = RuleSet()

    rule_set.from_tsv(os.path.join(dir_name, "features.tsv"))

    assert rule_set.to_list() == [FEATURE1, FEATURE2]


def test_ruleset_openie():
    rule_set = RuleSet([Rule(FEATURE3, openie=True), Rule(FEATURE4, openie=True)])
    rule_set.to_tsv(os.path.join(dir_name, "features_openie.tsv"))

    rule_set = RuleSet()
    rule_set.from_tsv(os.path.join(dir_name, "features_openie.tsv"))

    assert rule_set.to_list() == [FEATURE3, FEATURE4]


def test_openie_matching():
    evaluator = FeatureEvaluator()

    G, _ = default_pn_to_graph(GRAPH)

    rule_set = RuleSet([Rule(FEATURE3, openie=True), Rule(FEATURE4, openie=True)])

    triplets = list(evaluator.annotate(G, rule_set.to_list()))

    assert triplets == [
        {"relation": "FIND", "ARG1": "location", "ARG2": "object"},
        {"relation": "WRITE", "ARG1": "person", "ARG2": "sentence"},
    ]



from re import I
from typing import List, Tuple, Dict

import networkx as nx
import pandas as pd

from tqdm import tqdm
from tuw_nlp.graph.utils import graph_to_pn
from xpotato.dataset.sample import Sample
from xpotato.graph_extractor.extract import GraphExtractor
from xpotato.graph_extractor.graph import PotatoGraph


class Dataset:
    def __init__(
        self,
        examples: List[Tuple[str, str]] = None,
        label_vocab: Dict[str, int] = {},
        lang="en",
        path=None,
        binary=False,
    ) -> None:
        self.label_vocab = label_vocab
        if path:
            self._dataset = self.read_dataset(path=path, binary=binary)
        else:
            self._dataset = self.read_dataset(examples=examples)
        self.extractor = GraphExtractor(lang=lang, cache_fn=f"{lang}_nlp_cache")
        self.graphs = None

    @staticmethod
    def save_dataframe(df: pd.DataFrame, path: str) -> None:
        graphs = [graph_to_pn(graph) for graph in df["graph"].tolist()]
        df["graph"] = graphs
        df.to_csv(path, index=False, sep="\t")

    def prune_graphs(self, graphs: List[nx.DiGraph] = None) -> None:
        graphs_str = []
        for i, graph in enumerate(graphs):
            graph.remove_nodes_from(list(nx.isolates(graph)))
            # ADAM: THIS IS JUST FOR PICKLE TO PENMAN CONVERSION
            graph = self._random_postprocess(graph)

            g = [
                c
                for c in sorted(
                    nx.weakly_connected_components(graph), key=len, reverse=True
                )
            ]
            if len(g) > 1:
                print(
                    "WARNING: graph has multiple connected components, taking the largest"
                )
                g_pn = graph_to_pn(graph.subgraph(g[0].copy()))
            else:
                g_pn = graph_to_pn(graph)

            graphs_str.append(g_pn)

        return graphs_str

    def read_dataset(
        self,
        examples: List[Tuple[str, str]] = None,
        path: str = None,
        binary: bool = False,
    ) -> List[Sample]:
        if examples:
            return [Sample(example, PotatoGraph()) for example in examples]
        elif path:
            if binary:
                df = pd.read_pickle(path)
                graphs_str = self.prune_graphs(df.graph.tolist())
                df.drop(columns=["graph"], inplace=True)
                df["graph"] = graphs_str
            else:
                df = pd.read_csv(path, sep="\t")

            return [
                Sample(
                    (example["text"], example["label"]),
                    potato_graph=PotatoGraph(graph_str=example["graph"]),
                    label_id=example["label_id"],
                )
                for _, example in tqdm(df.iterrows())
            ]
        else:
            raise ValueError("No examples or path provided")

    # ADAM: THIS WILL NEED TO BE ADDRESSED
    def _random_postprocess(self, graph: nx.DiGraph) -> nx.DiGraph:
        for node, attr in graph.nodes(data=True):
            if len(attr["name"].split()) > 1:
                attr["name"] = attr["name"].split()[0]

        return graph

    def to_dataframe(self) -> pd.DataFrame:
        df = pd.DataFrame(
            {
                "text": [sample.text for sample in self._dataset],
                "label": [sample.label for sample in self._dataset],
                "label_id": [
                    sample.get_label_id(self.label_vocab) for sample in self._dataset
                ],
                "graph": [sample.potato_graph.graph for sample in self._dataset],
            }
        )
        return df

    def parse_graphs(self, graph_format: str = "fourlang") -> List[nx.DiGraph]:
        graphs = list(
            self.extractor.parse_iterable(
                [sample.text for sample in self._dataset], graph_format
            )
        )

        self.graphs = [PotatoGraph(graph) for graph in graphs]
        return self.graphs

    def set_graphs(self, graphs: List[PotatoGraph]) -> None:
        for sample, potato_graph in zip(self._dataset, graphs):
            potato_graph.graph.remove_edges_from(nx.selfloop_edges(potato_graph.graph))
            sample.set_graph(potato_graph)

    def load_graphs(self, path: str, binary: bool = False) -> None:
        if binary:
            graphs = [graph for graph in pd.read_pickle(path)]
            graph_str = self.prune_graphs(graphs)

            graphs = [PotatoGraph(graph_str=graph) for graph in graph_str]
            self.graphs = graphs
        else:
            with open(path, "rb") as f:
                for line in f:
                    graph = PotatoGraph(graph_str=line.strip())
                    self.graphs.append(graph)

        self.set_graphs(self.graphs)

    def save_dataset(self, path: str) -> None:
        df = pd.DataFrame(
            {
                "text": [sample.text for sample in self._dataset],
                "label": [sample.label for sample in self._dataset],
                "label_id": [
                    sample.get_label_id(self.label_vocab) for sample in self._dataset
                ],
                "graph": [str(sample.potato_graph) for sample in self._dataset],
            }
        )
        df.to_csv(path, index=False, sep="\t")

    def save_graphs(self, path: str) -> None:
        with open(path, "wb") as f:
            for graph in self.graphs:
                f.write(str(graph) + "\n")



import argparse
import json
import logging
import sys

import pandas as pd
from sklearn.metrics import classification_report

from xpotato.graph_extractor.extract import FeatureEvaluator
from xpotato.graph_extractor.graph import PotatoGraph


# TODO Adam: This is not the best place for these functions but I didn't want it to be in the frontend.utils
# ------------------------------------------------------


def filter_label(df, label):
    df["label"] = df.apply(lambda x: label if label in x["labels"] else "NOT", axis=1)
    df["label_id"] = df.apply(lambda x: 0 if x["label"] == "NOT" else 1, axis=1)


def read_df(path, label=None, binary=False):
    if binary:
        df = pd.read_pickle(path)
    else:
        df = pd.read_csv(path, sep="\t")
        graphs = []
        for graph in df["graph"]:
            potato_graph = PotatoGraph(graph_str=graph)
            graphs.append(potato_graph.graph)
        df["graph"] = graphs
    if label is not None:
        filter_label(df, label)
    return df


# ------------------------------------------------------


def get_args():
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("-t", "--graph-type", type=str, default="fourlang")
    parser.add_argument("-f", "--features", type=str, required=True)
    parser.add_argument("-d", "--dataset-path", type=str, default=None, required=True)
    parser.add_argument("-m", "--mode", type=str, default="predictions")
    parser.add_argument(
        "-l",
        "--label",
        default=None,
        type=str,
        help="Specify label for OneVsAll multi-label classification. Datasets require a labels column with all valid labels.",
    )
    return parser.parse_args()


def main():
    logging.basicConfig(
        level=logging.WARNING,
        format="%(asctime)s : "
        + "%(module)s (%(lineno)s) - %(levelname)s - %(message)s",
    )

    args = get_args()
    df = read_df(args.dataset_path, args.label)

    with open(args.features) as f:
        features = json.load(f)

    feature_values = []
    for k in features:
        for f in features[k]:
            feature_values.append(f)
    evaluator = FeatureEvaluator()
    pred_df = evaluator.match_features(df, feature_values)
    report = None
    if "label" in df and df["label"].iloc[0]:
        pred_df["label"] = df.label

        label_to_id = {}

        for label in df.groupby("label").size().keys().tolist():
            if label in label_to_id:
                continue
            else:
                label_to_id[label] = df[df.label == label].iloc[1].label_id

        predicted_label = []
        gold = df.label_id.tolist()

        for label in pred_df["Predicted label"]:
            if label in label_to_id:
                predicted_label.append(label_to_id[label])
            else:
                predicted_label.append(0)

        report = classification_report(gold, predicted_label, digits=3)

    if args.mode == "predictions":
        pred_df.to_csv(sys.stdout, sep="\t")
    elif args.mode == "report":
        assert (
            report
        ), "There are no labels in the dataset, we cannot generate a classification report. Are you evaluating a test set?"
        print(report)


if __name__ == "__main__":
    main()



from typing import Dict, Tuple

from xpotato.graph_extractor.graph import PotatoGraph


class Sample:
    def __init__(
        self,
        example: Tuple[str, str],
        potato_graph: PotatoGraph = None,
        label_id: int = None,
    ) -> None:
        self.text = example[0]
        self.label = example[1]
        self.label_id = label_id
        self.potato_graph = potato_graph

    def set_graph(self, graph: PotatoGraph) -> None:
        self.potato_graph = graph

    def get_label_id(self, label_vocab: Dict[str, int]):
        if self.label_id is not None:
            return self.label_id
        elif self.label and self.label in label_vocab:
            return label_vocab[self.label]
        else:
            return None



import argparse
import json
import logging
import sys

from xpotato.dataset.dataset import Dataset
from xpotato.dataset.utils import save_dataframe


def get_args():
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("-p", "--pickle", type=str, required=True)
    parser.add_argument("-o", "--output", type=str, required=True)
    return parser.parse_args()


def main():
    args = get_args()

    path = args.pickle
    output = args.output

    dataset = Dataset(path=path, binary=True)
    df = dataset.to_dataframe()

    save_dataframe(df, output)


if __name__ == "__main__":
    main()



from xpotato.graph_extractor.extract import GraphExtractor
from xpotato.graph_extractor.extract import FeatureEvaluator
from xpotato.dataset.utils import default_pn_to_graph
from tuw_nlp.graph.utils import (
    GraphFormulaMatcher,
)

import json
import os

from fastapi import FastAPI
from pydantic import BaseModel


FEATURE_PATH = os.getenv("FEATURE_PATH", None)
GRAPH_FORMAT = os.getenv("GRAPH_FORMAT", None)
LANG = os.getenv("LANG", None)

assert FEATURE_PATH, "FEATURE_PATH is not set"
assert GRAPH_FORMAT, "GRAPH_FORMAT is not set"
assert LANG, "LANG is not set"


def match_texts(text):
    texts = text.split("\n")

    graphs = list(EXTRACTOR.parse_iterable([text for text in texts], GRAPH_FORMAT))

    predicted = []

    for i, g in enumerate(graphs):
        feats = MATCHER.match(g)
        label = "NONE"
        for key, feature in feats:
            label = key
        predicted.append(label)

    return predicted


class Item(BaseModel):
    text: str


app = FastAPI()


@app.on_event("startup")
def init_data():
    global EXTRACTOR
    global FEATURES
    global EVALUATOR
    global MATCHER
    EXTRACTOR = GraphExtractor(lang=LANG, cache_fn=f"{LANG}_nlp_cache")

    if GRAPH_FORMAT == "ud":
        EXTRACTOR.init_nlp()
    elif GRAPH_FORMAT == "amr":
        EXTRACTOR.init_amr()

    with open(FEATURE_PATH) as f:
        FEATURES = json.load(f)

    EVALUATOR = FeatureEvaluator()

    feature_values = []
    for k in FEATURES:
        for f in FEATURES[k]:
            feature_values.append(f)
    MATCHER = GraphFormulaMatcher(feature_values, converter=default_pn_to_graph)


@app.post("/")
async def infer(item: Item):
    predicted = match_texts(text=item.text)
    return predicted


if __name__ == "__main__":
    import uvicorn

    uvicorn.run("main:app", port=8000, reload=True)



import argparse
import copy
import os
import time
import json
import streamlit as st
import pandas as pd
import penman
from graphviz import Source

from st_aggrid import AgGrid, GridOptionsBuilder, GridUpdateMode
from tuw_nlp.graph.utils import graph_to_pn

from utils import (
    train_df,
    add_rule_manually,
    annotate_df,
    extract_data_from_dataframe,
    get_df_from_rules,
    graph_viewer,
    init_evaluator,
    init_extractor,
    init_session_states,
    rank_and_suggest,
    read_df,
    rerun,
    rule_chooser,
    save_ruleset,
    read_ruleset,
    save_after_modify,
    save_dataframe,
    match_texts,
    show_ml_feature,
    st_stdout,
    to_dot,
)


def inference_mode(evaluator, hand_made_rules):
    st.sidebar.title("Inference Mode")
    st.sidebar.markdown(
        """
    ## Select the graph format
    """
    )
    graph_format = st.sidebar.selectbox("", ["ud", "fourlang", "amr"])
    st.sidebar.markdown(
        """
    ## Select the language
    """
    )
    lang = st.sidebar.selectbox("", ["en", "de"])
    if lang == "de" and graph_format == "amr":
        st.error("Currently AMR does not support German")
        return

    st.sidebar.markdown(
        """
        __Show download button for graphs (an option to allow downloading graphs)__
        """
    )
    st.session_state.download = st.sidebar.selectbox("", options=[False, True], key=2)

    if hand_made_rules:
        read_ruleset(hand_made_rules)

    extractor = init_extractor(lang, graph_format)

    col1, col2 = st.columns(2)

    col1.header("Rule chooser and modifier")

    col2.header("Graph viewer and evaluator")

    with col1:
        features_merged = []
        for i in st.session_state.features:
            for j in st.session_state.features[i]:
                features_merged.append(j)

        st.session_state.feature_df = get_df_from_rules(
            [";".join(feat[0]) for feat in features_merged],
            [";".join(feat[1]) for feat in features_merged],
            [feat[2] for feat in features_merged],
        )

        with st.form("example form") as f:
            gb = GridOptionsBuilder.from_dataframe(st.session_state.feature_df)
            # make all columns editable
            gb.configure_columns(
                ["rules", "negated_rules", "predicted_label"], editable=True
            )
            gb.configure_selection(
                "multiple",
                use_checkbox=True,
                groupSelectsChildren=True,
                groupSelectsFiltered=True,
            )
            go = gb.build()
            ag = AgGrid(
                st.session_state.feature_df,
                gridOptions=go,
                key="grid1",
                allow_unsafe_jscode=True,
                reload_data=True,
                update_mode=GridUpdateMode.MODEL_CHANGED | GridUpdateMode.VALUE_CHANGED,
                width="100%",
                theme="light",
                fit_columns_on_grid_load=False,
            )

            delete_or_train = st.radio("Delete selected rules", ("none", "delete"))
            submit = st.form_submit_button(label="save updates")

        if submit:
            delete = delete_or_train == "delete"

            st.session_state.rows_to_delete = [r["rules"] for r in ag["selected_rows"]]
            st.session_state.rls_after_delete = []

            negated_list = ag["data"]["negated_rules"].tolist()
            predicted_labels = ag["data"]["predicted_label"].tolist()
            feature_list = []
            for i, rule in enumerate(ag["data"]["rules"].tolist()):
                if not negated_list[i].strip():
                    feature_list.append([rule.split(";"), [], predicted_labels[i]])
                else:
                    feature_list.append(
                        [
                            rule.split(";"),
                            negated_list[i].strip().split(";"),
                            predicted_labels[i],
                        ]
                    )
            if st.session_state.rows_to_delete and delete:
                for r in feature_list:
                    if ";".join(r[0]) not in st.session_state.rows_to_delete:
                        st.session_state.rls_after_delete.append(r)
            else:
                st.session_state.rls_after_delete = copy.deepcopy(feature_list)

            if st.session_state.rls_after_delete and not delete:
                save_after_modify(hand_made_rules)

        if st.session_state.rows_to_delete and delete_or_train == "delete":
            with st.form("Delete form"):
                st.write("The following rules will be deleted, do you accept it?")
                st.write(st.session_state.rows_to_delete)
                save_button = st.form_submit_button("Accept Delete")

            if save_button:
                save_after_modify(hand_made_rules)
        text = st.text_area("You can add a new rule here manually")
        negated_text = st.text_area("You can modify the negated features here")
        label_text = st.text_area("You can modify the label here")
        agree = st.button("Add rule to the ruleset")
        if agree:
            if not negated_text.strip():
                negated_features = []
            else:
                negated_features = negated_text.split(";")
            if not label_text.strip():
                with st_stdout("error"):
                    print("Please provide label for your rule!")
            else:
                if label_text.strip() not in st.session_state.features:
                    st.session_state.features[label_text.strip()] = []

                st.session_state.features[label_text].append(
                    [[text], negated_features, label_text]
                )
                if st.session_state.features[label_text]:
                    features_merged = []
                    for i in st.session_state.features:
                        for j in st.session_state.features[i]:
                            features_merged.append(j)

                    st.session_state.feature_df = get_df_from_rules(
                        [";".join(feat[0]) for feat in features_merged],
                        [";".join(feat[1]) for feat in features_merged],
                        [feat[2] for feat in features_merged],
                    )
                    save_rules = hand_made_rules or "saved_features.tsv"
                    save_ruleset(save_rules, st.session_state.features)
                    rerun()

    with col2:
        text_input = st.text_area(
            "Provide the text here you want to parse and apply rules to!"
        )

        if text_input:
            graphs, predicted = match_texts(text_input, extractor, graph_format)
            dot_current_graph = to_dot(
                graphs[0],
            )

            st.markdown("__The predicted labels are:__")
            st.write(predicted[0][0])

            st.markdown("__The matched patterns are:__")
            st.write(predicted[0][1])
            if st.session_state.download:
                graph_pipe = Source(dot_current_graph).pipe(format="svg")
                st.download_button(
                    label="Download graph as SVG",
                    data=graph_pipe,
                    file_name="graph.svg",
                    mime="mage/svg+xml",
                )

            with st.expander("Graph dot source", expanded=False):
                st.write(dot_current_graph)

            st.graphviz_chart(
                dot_current_graph,
                use_container_width=True,
            )

            st.write("Penman format:")
            st.text(penman.encode(penman.decode(graph_to_pn(graphs[0])), indent=10))
            st.write("In one line format:")
            st.write(graph_to_pn(graphs[0]))


def simple_mode(evaluator, data, val_data, graph_format, feature_path, hand_made_rules):
    if hand_made_rules:
        read_ruleset(hand_made_rules)

    if "df" not in st.session_state:
        st.session_state.df = data.copy()
        if "index" not in st.session_state.df:
            # First reset the index, so it starts from 0 and increments sequentially
            # Then reset again and add the index as a column
            st.session_state.df.reset_index(level=0, inplace=True, drop=True)
            st.session_state.df.reset_index(level=0, inplace=True)

    if not feature_path and not st.session_state.trained:
        st.sidebar.title("Train your dataset!")
        show_app = st.sidebar.button("Train")
        st.sidebar.markdown(
            """
        __Min edge in features (if you set it higher than 0, we won't allow single tokens in the features, only edges)__
        """
        )
        st.session_state.min_edge = st.sidebar.number_input(
            "", min_value=0, max_value=3, value=0, step=1
        )
        st.sidebar.markdown(
            """
            __Ranking type:__
             - False -> train a machine learning model and rank based on feature importance
             - True -> rank all features based on TP and FP scores
            """
        )
        st.session_state.rank = st.sidebar.selectbox("", options=[False, True], key=1)

        st.sidebar.markdown(
            """
        __Show download button for graphs (an option to allow downloading graphs)__
        """
        )
        st.session_state.download = st.sidebar.selectbox(
            "", options=[False, True], key=2
        )
        if show_app:
            st.session_state.suggested_features = train_df(
                data, st.session_state.min_edge
            )
            st.session_state.trained = True
            with st_stdout("success"):
                print("Success, your dataset is trained, wait for the app to load..")
                time.sleep(3)
                rerun()
        st.markdown(
            "<h3 style='text-align: center; color: black;'>Your dataset is shown below, click the train button to train your dataset!</h3>",
            unsafe_allow_html=True,
        )
        sample_df = AgGrid(data, width="100%", fit_columns_on_grid_load=True)

        st.write("label distribution:")
        st.bar_chart(data.groupby("label").size())

        st.write("sentence lenghts:")
        st.bar_chart(data.text.str.len().sort_values(ascending=False).iloc[0:100])

        st.write("common words:")
        st.bar_chart(
            pd.Series(" ".join(data["text"]).lower().split()).value_counts()[:100]
        )

    if st.session_state.trained or feature_path:

        with st.expander("Browse dataset:"):
            st.markdown(
                """
                The dataset can be inspected below, if you are interested in of the graphs, you can check it in the graph browser under 'Graphs and results'
                """
            )
            gb = GridOptionsBuilder.from_dataframe(st.session_state.df)
            gb.configure_default_column(
                editable=False,
                resizable=True,
                sorteable=True,
                wrapText=True,
                autoHeight=True,
            )
            gb.configure_column("graph", hide=True)

            go = gb.build()
            selected_df = AgGrid(
                st.session_state.df,
                gridOptions=go,
                allow_unsafe_jscode=True,
                reload_data=False,
                update_mode=GridUpdateMode.MODEL_CHANGED | GridUpdateMode.VALUE_CHANGED,
                width="100%",
                theme="light",
                fit_columns_on_grid_load=True,
            )

        col1, col2 = st.columns(2)

        if (
            feature_path
            and os.path.exists(feature_path)
            and not st.session_state.suggested_features
        ):
            with open(feature_path) as f:
                st.session_state.suggested_features = json.load(f)

        if not st.session_state.features:
            for key in st.session_state.suggested_features:
                pop_len = (
                    5
                    if len(st.session_state.suggested_features[key]) > 5
                    else len(st.session_state.suggested_features[key])
                )
                st.session_state.features[key] = [
                    st.session_state.suggested_features[key].pop(0)
                    for _ in range(pop_len)
                ]

        col1.header("Rule chooser and modifier")

        col2.header("Graph viewer and evaluator")

        with col1:
            classes = st.selectbox(
                "First, choose class you want to use to build rules",
                list(st.session_state.features.keys()),
            )

            st.session_state.feature_df = get_df_from_rules(
                [";".join(feat[0]) for feat in st.session_state.features[classes]],
                [";".join(feat[1]) for feat in st.session_state.features[classes]],
            )

            with st.form("example form") as f:

                st.markdown(
                    """
                __You can modify any rule you want to__

                Remember, we use the [PENMAN](https://github.com/goodmami/penman) notation to describe a rule. You can find more information about the rules in the [README](https://github.com/adaamko/POTATO) of our repository.
                """
                )

                gb = GridOptionsBuilder.from_dataframe(st.session_state.feature_df)
                # make all columns editable
                gb.configure_columns(["rules", "negated_rules"], editable=True)
                gb.configure_selection(
                    "multiple",
                    use_checkbox=True,
                    groupSelectsChildren=True,
                    groupSelectsFiltered=True,
                )
                go = gb.build()
                ag = AgGrid(
                    st.session_state.feature_df,
                    gridOptions=go,
                    key="grid1",
                    allow_unsafe_jscode=True,
                    reload_data=True,
                    update_mode=GridUpdateMode.MODEL_CHANGED
                    | GridUpdateMode.VALUE_CHANGED,
                    width="100%",
                    theme="light",
                    fit_columns_on_grid_load=False,
                )

                st.markdown(
                    """
                    After you modified any rule, click on __save updates__ button to save your changes.


                    Tick the __box__ next to the rules you want to evaluate and then, click on the __evaluate__ button.
                    If you don't tick any rule, the evaluator will evaluate all the rules.
                """
                )
                delete_or_train = st.radio(
                    "Delete or Refine selected rules (you can only refine underspecified rules that contain a regexp)",
                    ("none", "delete", "refine"),
                )
                submit = st.form_submit_button(label="save updates")
                evaluate = st.form_submit_button(label="evaluate selected")

            if evaluate:
                feature_list = []
                selected_rules = (
                    ag["selected_rows"]
                    if ag["selected_rows"]
                    else ag["data"].to_dict(orient="records")
                )
                for rule in selected_rules:
                    positive_rules = (
                        rule["rules"].split(";")
                        if "rules" in rule and rule["rules"].strip()
                        else []
                    )
                    negated_rules = (
                        rule["negated_rules"].split(";")
                        if "negated_rules" in rule and rule["negated_rules"].strip()
                        else []
                    )
                    feature_list.append(
                        [
                            positive_rules,
                            negated_rules,
                            classes,
                        ]
                    )
                st.session_state.sens = [";".join(feat[0]) for feat in feature_list]
                with st.spinner("Evaluating rules..."):
                    (
                        st.session_state.df_statistics,
                        st.session_state.whole_accuracy,
                    ) = evaluator.evaluate_feature(
                        classes, feature_list, data, graph_format
                    )
                    (
                        st.session_state.val_dataframe,
                        st.session_state.whole_accuracy_val,
                    ) = evaluator.evaluate_feature(
                        classes,
                        feature_list,
                        val_data,
                        graph_format,
                    )
                st.success("Done!")
                rerun()

            if submit:
                delete = delete_or_train == "delete"
                train = delete_or_train == "refine"

                st.session_state.rows_to_delete = [
                    r["rules"] for r in ag["selected_rows"]
                ]
                st.session_state.rls_after_delete = []

                feature_list = []
                selected_rules = ag["data"].to_dict(orient="records")

                for rule in selected_rules:
                    positive_rules = (
                        rule["rules"].split(";")
                        if "rules" in rule and rule["rules"].strip()
                        else []
                    )
                    negated_rules = (
                        rule["negated_rules"].split(";")
                        if "negated_rules" in rule and rule["negated_rules"].strip()
                        else []
                    )
                    feature_list.append(
                        [
                            positive_rules,
                            negated_rules,
                            classes,
                        ]
                    )
                if st.session_state.rows_to_delete and delete:
                    for r in feature_list:
                        if ";".join(r[0]) not in st.session_state.rows_to_delete:
                            st.session_state.rls_after_delete.append(r)
                elif st.session_state.rows_to_delete and train:
                    st.session_state.rls_after_delete = copy.deepcopy(feature_list)
                    rule_to_train = st.session_state.rows_to_delete[0]
                    if ";" in rule_to_train or ".*" not in rule_to_train:
                        st.text("Only single and underspecified rules can be refined!")
                    else:
                        selected_words = evaluator.train_feature(
                            classes, rule_to_train, data, graph_format
                        )

                        for f in selected_words:
                            st.session_state.rls_after_delete.append([[f], [], classes])
                else:
                    st.session_state.rls_after_delete = copy.deepcopy(feature_list)

                if st.session_state.rls_after_delete and not delete:
                    save_after_modify(hand_made_rules, classes)

            if st.session_state.rows_to_delete and delete_or_train == "delete":
                with st.form("Delete form"):
                    st.write("The following rules will be deleted, do you accept it?")
                    st.write(st.session_state.rows_to_delete)
                    save_button = st.form_submit_button("Accept Delete")

                if save_button:
                    save_after_modify(hand_made_rules, classes)

            add_rule_manually(classes, hand_made_rules)
            rank_and_suggest(classes, data, evaluator)

            if st.session_state.ml_feature:
                show_ml_feature(classes, hand_made_rules)

        with col2:
            # THIS IS HERE BECAUSE STREAMLIT IS BUGGY AND DOESN'T DISPLAY TWO DOT GRAPHS IN DIFFERENT CONTAINERS
            # NEXT RELEASE WILL FIX IT
            with st.expander("Browse graphs:"):
                graph_id = st.number_input(
                    label="Input the ID of the graph you want to view", min_value=0
                )

                browse_current_graph_nx = (
                    st.session_state.df[st.session_state.df.index == graph_id]
                    .iloc[0]
                    .graph
                )
                browse_current_graph = to_dot(browse_current_graph_nx)
                if st.session_state.download:
                    graph_pipe = Source(browse_current_graph).pipe(format="svg")
                    st.download_button(
                        label="Download graph as SVG",
                        data=graph_pipe,
                        file_name="graph.svg",
                        mime="mage/svg+xml",
                    )

                st.graphviz_chart(
                    browse_current_graph,
                    use_container_width=True,
                )

                st.write("Penman format:")
                st.text(
                    penman.encode(
                        penman.decode(graph_to_pn(browse_current_graph_nx)), indent=10
                    )
                )
                st.write("In one line format:")
                st.write(graph_to_pn(browse_current_graph_nx))
            if not st.session_state.df_statistics.empty and st.session_state.sens:
                if st.session_state.sens:
                    nodes, option = rule_chooser()
                st.markdown(
                    f"<span>Result of using all the rules: Precision: <b>{st.session_state.whole_accuracy[0]:.3f}</b>, \
                        Recall: <b>{st.session_state.whole_accuracy[1]:.3f}</b>, Fscore: <b>{st.session_state.whole_accuracy[2]:.3f}</b></span>",
                    unsafe_allow_html=True,
                )
                (
                    fn_graphs,
                    fn_sentences,
                    fn_indices,
                    fp_graphs,
                    fp_sentences,
                    fp_indices,
                    fscore,
                    prec,
                    predicted,
                    recall,
                    support,
                    tp_graphs,
                    tp_sentences,
                    tp_indices,
                ) = extract_data_from_dataframe(option)

                st.markdown(
                    f"<span>The rule's result: Precision: <b>{prec:.3f}</b>, Recall: <b>{recall:.3f}</b>, \
                        Fscore: <b>{fscore:.3f}</b>, True positives: <b>{len(tp_graphs)}</b>, False positives: <b>{len(fp_graphs)}</b></span>",
                    unsafe_allow_html=True,
                )

                with st.expander("Show validation data", expanded=False):
                    val_prec = st.session_state.val_dataframe.iloc[
                        st.session_state.sens.index(option)
                    ].Precision
                    val_recall = st.session_state.val_dataframe.iloc[
                        st.session_state.sens.index(option)
                    ].Recall
                    val_fscore = st.session_state.val_dataframe.iloc[
                        st.session_state.sens.index(option)
                    ].Fscore
                    val_support = st.session_state.val_dataframe.iloc[
                        st.session_state.sens.index(option)
                    ].Support
                    st.markdown(
                        f"<span>Result of using all the rules on the validation data: Precision: <b>{st.session_state.whole_accuracy_val[0]:.3f}</b>, \
                            Recall: <b>{st.session_state.whole_accuracy_val[1]:.3f}</b>, Fscore: <b>{st.session_state.whole_accuracy_val[2]:.3f}</b>, \
                                Support: <b>{st.session_state.whole_accuracy_val[3]}</b></span>",
                        unsafe_allow_html=True,
                    )
                    st.markdown(
                        f"<span>The rule's result on the validation data: Precision: <b>{val_prec:.3f}</b>, \
                            Recall: <b>{val_recall:.3f}</b>, Fscore: <b>{val_fscore:.3f}</b>, \
                                Support: <b>{val_support}</b></span>",
                        unsafe_allow_html=True,
                    )

                tp_fp_fn_choice = (
                    "True Positive graphs",
                    "False Positive graphs",
                    "False Negative graphs",
                )
                tp_fp_fn = st.selectbox(
                    "Select the graphs you want to view", tp_fp_fn_choice
                )
                if tp_fp_fn == "False Positive graphs":
                    if fp_graphs:
                        graph_viewer("FP", fp_graphs, fp_sentences, fp_indices, nodes)

                elif tp_fp_fn == "True Positive graphs":
                    if tp_graphs:
                        graph_viewer("TP", tp_graphs, tp_sentences, tp_indices, nodes)

                elif tp_fp_fn == "False Negative graphs":
                    if fn_graphs:
                        graph_viewer("FN", fn_graphs, fn_sentences, fn_indices, nodes)


def advanced_mode(evaluator, train_data, graph_format, feature_path, hand_made_rules):
    data = read_df(train_data)
    if hand_made_rules:
        read_ruleset(hand_made_rules)
    if "df" not in st.session_state:
        st.session_state.df = data.copy()
        if "annotated" not in st.session_state.df:
            st.session_state.df["annotated"] = False
        if "applied_rules" not in st.session_state.df:
            st.session_state.df["applied_rules"] = [
                [] for _ in range(len(st.session_state.df))
            ]
        if "index" not in st.session_state.df:
            # First reset the index, so it starts from 0 and increments sequentially
            # Then reset again and add the index as a column
            st.session_state.df.reset_index(level=0, inplace=True, drop=True)
            st.session_state.df.reset_index(level=0, inplace=True)

    df_annotated = st.session_state.df[st.session_state.df.annotated == True][
        ["index", "text", "label", "applied_rules"]
    ]
    df_unannotated = st.session_state.df[st.session_state.df.annotated == False][
        ["index", "text", "label", "applied_rules"]
    ]

    # First we need to provide the labels we want to annotate
    if "labels" not in st.session_state:
        st.text("Before we start, please provide labels you want to train")
        user_input = st.text_input("label encoding", placeholder="NOT:0,OFF:1")

        if st.button("Add labels"):
            if user_input:
                try:
                    labs = user_input.split(",")
                    assert (
                        len(labs) == 2
                    ), "Please provide only two labels, currently we only support binary annotations!"
                    st.session_state.labels = {
                        label.split(":")[0]: int(label.split(":")[1]) for label in labs
                    }

                    st.write(st.session_state.labels)
                    st.session_state.inverse_labels = {
                        v: k for (k, v) in st.session_state.labels.items()
                    }
                    rerun()
                except Exception as e:
                    st.write(e)
                    st.write("Bad format, the right format is NOT:0,OFF:1")
            else:
                st.write("No labels provided!")

    else:
        with st.expander("Annotation/Dataset browser:"):
            st.markdown(
                f"<span><b>Annotate samples here:</b></span>",
                unsafe_allow_html=True,
            )

            if st.session_state.applied_rules:
                st.markdown(
                    f"<span>Currently the following rules are applied:</span>",
                    unsafe_allow_html=True,
                )
                st.write(st.session_state.applied_rules)
            with st.form("annotate form") as f:
                gb = GridOptionsBuilder.from_dataframe(df_unannotated)
                gb.configure_default_column(
                    editable=True,
                    resizable=True,
                    sorteable=True,
                    wrapText=True,
                    autoHeight=True,
                )
                # make all columns editable
                gb.configure_selection(
                    "multiple",
                    use_checkbox=True,
                    groupSelectsChildren=True,
                    groupSelectsFiltered=True,
                )
                go = gb.build()
                ag = AgGrid(
                    df_unannotated,
                    gridOptions=go,
                    key="grid2",
                    allow_unsafe_jscode=True,
                    reload_data=True,
                    update_mode=GridUpdateMode.MODEL_CHANGED
                    | GridUpdateMode.VALUE_CHANGED,
                    width="100%",
                    theme="light",
                    fit_columns_on_grid_load=True,
                )

                annotate = st.form_submit_button("Annotate")

            if annotate:
                if ag["selected_rows"]:
                    for row in ag["selected_rows"]:
                        st.session_state.df.loc[
                            row["index"], "label"
                        ] = st.session_state.inverse_labels[1]
                        st.session_state.df.loc[row["index"], "annotated"] = True
                    save_dataframe(st.session_state.df, train_data)
                    rerun()

            st.markdown(
                f"<span>Samples you have already annotated:</span>",
                unsafe_allow_html=True,
            )
            with st.form("annotated form") as f:
                gb = GridOptionsBuilder.from_dataframe(df_annotated)
                gb.configure_default_column(
                    editable=True,
                    resizable=True,
                    sorteable=True,
                    wrapText=True,
                )
                # make all columns editable
                gb.configure_selection(
                    "multiple",
                    use_checkbox=True,
                    groupSelectsChildren=True,
                    groupSelectsFiltered=True,
                )
                go = gb.build()
                ag_ann = AgGrid(
                    df_annotated,
                    gridOptions=go,
                    key="grid3",
                    allow_unsafe_jscode=True,
                    reload_data=True,
                    update_mode=GridUpdateMode.MODEL_CHANGED
                    | GridUpdateMode.VALUE_CHANGED,
                    width="100%",
                    theme="light",
                    fit_columns_on_grid_load=True,
                )

                clear_annotate = st.form_submit_button("Clear annotation")

            if clear_annotate:
                if ag_ann["selected_rows"]:
                    for row in ag_ann["selected_rows"]:
                        st.session_state.df.loc[
                            row["index"], "label"
                        ] = st.session_state.inverse_labels[1]
                        st.session_state.df.loc[row["index"], "annotated"] = False
                        st.session_state.df.loc[row["index"], "label"] = ""
                    save_dataframe(st.session_state.df, train_data)
                    rerun()

        st.sidebar.title("Train your dataset!")
        train = st.sidebar.button("Train!")
        st.sidebar.markdown(
            """
        __Min edge in features (if you set it higher than 0, we won't allow single tokens in the features, only edges)__
        """
        )
        st.session_state.min_edge = st.sidebar.number_input(
            "", min_value=0, max_value=3, value=0, step=1
        )
        st.sidebar.markdown(
            """
            __Ranking type:__
             - False -> train a machine learning model and rank based on feature importance
             - True -> rank all features based on TP and FP scores
            """
        )
        st.session_state.rank = st.sidebar.selectbox("", options=[False, True], key=1)

        st.sidebar.markdown(
            """
        __Show download button for graphs (an option to allow downloading graphs)__
        """
        )
        st.session_state.download = st.sidebar.selectbox(
            "", options=[False, True], key=2
        )
        if train:
            df_to_train = st.session_state.df.copy()
            # df_to_train = df_to_train[df_to_train.applied_rules.map(len) == 0]

            if not df_to_train.empty:
                st.session_state.trained = True
                df_to_train["label"] = df_to_train["label"].apply(
                    lambda x: st.session_state.inverse_labels[0] if not x else x
                )
                df_to_train["label_id"] = df_to_train["label"].apply(
                    lambda x: st.session_state.labels[x]
                )

                positive_size = df_to_train.groupby("label").size()[
                    st.session_state.inverse_labels[1]
                ]
                df_to_train = df_to_train.groupby("label").sample(
                    n=positive_size, random_state=1, replace=True
                )
                st.session_state.suggested_features = train_df(
                    df_to_train, st.session_state.min_edge, st.session_state.rank
                )
                st.session_state.df_to_train = df_to_train
                st.session_state.df_statistics = pd.DataFrame

            else:
                st.write("Empty dataframe!")

        col1, col2 = st.columns(2)

        with col1:
            if not st.session_state.features:
                st.session_state.features[st.session_state.inverse_labels[1]] = []

            classes = st.selectbox(
                "First, choose class you want to use to build rules",
                list(st.session_state.features.keys()),
            )

            if not st.session_state.features[classes]:
                st.session_state.features[classes] = [
                    [["(u_1 / dummy :1 (u_2 / feature))"], [], classes]
                ]
            st.session_state.feature_df = get_df_from_rules(
                [";".join(feat[0]) for feat in st.session_state.features[classes]],
                [";".join(feat[1]) for feat in st.session_state.features[classes]],
            )

            with st.form("example form") as f:
                st.markdown(
                    """
                __You can modify any rule you want to__

                Remember, we use the [PENMAN](https://github.com/goodmami/penman) notation to describe a rule. You can find more information about the rules in the [README](https://github.com/adaamko/POTATO) of our repository.
                """
                )

                gb = GridOptionsBuilder.from_dataframe(st.session_state.feature_df)
                # make all columns editable
                gb.configure_columns(["rules", "negated_rules"], editable=True)
                gb.configure_selection(
                    "multiple",
                    use_checkbox=True,
                    groupSelectsChildren=True,
                    groupSelectsFiltered=True,
                )
                go = gb.build()
                ag = AgGrid(
                    st.session_state.feature_df,
                    gridOptions=go,
                    key="grid1",
                    allow_unsafe_jscode=True,
                    reload_data=True,
                    update_mode=GridUpdateMode.MODEL_CHANGED
                    | GridUpdateMode.VALUE_CHANGED,
                    width="100%",
                    theme="light",
                    fit_columns_on_grid_load=True,
                )

                st.markdown(
                    """
                    After you modified any rule, click on __save updates__ button to save your changes.


                    Tick the __box__ next to the rules you want to evaluate and then, click on the __evaluate__ button.
                    If you don't tick any rule, the evaluator will evaluate all the rules.

                    If you think your rules are good and you want to evaluate samples based on them, click on the __annotate__ button.
                """
                )

                delete_or_train = st.radio(
                    "Delete or Refine selected rules (you can only refine underspecified rules that contain a regexp)",
                    ("none", "delete", "refine"),
                )
                submit = st.form_submit_button(label="save updates")
                evaluate = st.form_submit_button(label="evaluate selected")
                annotate = st.form_submit_button(label="annotate based on selected")

            if evaluate or annotate:
                feature_list = []
                selected_rules = (
                    ag["selected_rows"]
                    if ag["selected_rows"]
                    else ag["data"].to_dict(orient="records")
                )
                for rule in selected_rules:
                    positive_rules = (
                        rule["rules"].split(";")
                        if "rules" in rule and rule["rules"].strip()
                        else []
                    )
                    negated_rules = (
                        rule["negated_rules"].split(";")
                        if "negated_rules" in rule and rule["negated_rules"].strip()
                        else []
                    )
                    feature_list.append(
                        [
                            positive_rules,
                            negated_rules,
                            classes,
                        ]
                    )
                st.session_state.sens = [";".join(feat[0]) for feat in feature_list]
                if not st.session_state.sens:
                    with st_stdout("error"):
                        print(
                            "There are no rules to evaluate, please write some rules first, or get suggestions by our system!"
                        )

                else:
                    with st.spinner("Evaluating rules..."):
                        (
                            st.session_state.df_statistics,
                            st.session_state.whole_accuracy,
                        ) = evaluator.evaluate_feature(
                            classes,
                            feature_list,
                            st.session_state.df,
                            graph_format,
                        )

                    st.success("Done!")

                    if annotate:
                        predicted_rules = [[] for _ in range(len(st.session_state.df))]
                        st.session_state.applied_rules = st.session_state.sens
                        for j, opt in enumerate(st.session_state.sens):
                            predicted = st.session_state.df_statistics.iloc[j].Predicted

                            predicted_indices = [
                                i for i, pred in enumerate(predicted) if pred == 1
                            ]

                            for ind in predicted_indices:
                                predicted_rules[ind].append(opt)
                        annotate_df(predicted_rules)
                        st.session_state.trained = False

                    rerun()

            if submit:
                delete = delete_or_train == "delete"
                train = delete_or_train == "refine"

                st.session_state.rows_to_delete = [
                    r["rules"] for r in ag["selected_rows"]
                ]
                st.session_state.rls_after_delete = []

                feature_list = []
                selected_rules = ag["data"].to_dict(orient="records")

                for rule in selected_rules:
                    positive_rules = (
                        rule["rules"].split(";")
                        if "rules" in rule and rule["rules"].strip()
                        else []
                    )
                    negated_rules = (
                        rule["negated_rules"].split(";")
                        if "negated_rules" in rule and rule["negated_rules"].strip()
                        else []
                    )
                    feature_list.append(
                        [
                            positive_rules,
                            negated_rules,
                            classes,
                        ]
                    )
                if st.session_state.rows_to_delete and delete:
                    for r in feature_list:
                        if ";".join(r[0]) not in st.session_state.rows_to_delete:
                            st.session_state.rls_after_delete.append(r)
                elif st.session_state.rows_to_delete and train:
                    st.session_state.rls_after_delete = copy.deepcopy(feature_list)
                    rule_to_train = st.session_state.rows_to_delete[0]
                    if ";" in rule_to_train or ".*" not in rule_to_train:
                        st.text("Only single and underspecified rules can be refined!")
                    else:
                        selected_words = evaluator.train_feature(
                            classes,
                            rule_to_train,
                            st.session_state.df,
                            graph_format,
                        )

                        for f in selected_words:
                            st.session_state.rls_after_delete.append([[f], [], classes])
                else:
                    st.session_state.rls_after_delete = copy.deepcopy(feature_list)

                if st.session_state.rls_after_delete and not delete:
                    save_after_modify(hand_made_rules, classes)

            if st.session_state.rows_to_delete and delete_or_train == "delete":
                with st.form("Delete form"):
                    st.write("The following rules will be deleted, do you accept it?")
                    st.write(st.session_state.rows_to_delete)
                    save_button = st.form_submit_button("Accept Delete")

                if save_button:
                    save_after_modify(hand_made_rules, classes)

            add_rule_manually(classes, hand_made_rules)
            rank_and_suggest(classes, st.session_state.df, evaluator)

            if st.session_state.ml_feature:
                show_ml_feature(classes, hand_made_rules)
        with col2:
            with st.expander("Browse dataset/graphs"):
                graph_id = st.number_input(
                    label="Input the ID of the graph you want to view", min_value=0
                )

                browse_current_graph_nx = (
                    st.session_state.df[st.session_state.df.index == graph_id]
                    .iloc[0]
                    .graph
                )
                browse_current_graph = to_dot(browse_current_graph_nx)
                if st.session_state.download:
                    graph_pipe = Source(browse_current_graph).pipe(format="svg")
                    st.download_button(
                        label="Download graph as SVG",
                        data=graph_pipe,
                        file_name="graph.svg",
                        mime="mage/svg+xml",
                    )

                st.graphviz_chart(
                    browse_current_graph,
                    use_container_width=True,
                )
                st.write("Penman format:")
                st.text(
                    penman.encode(
                        penman.decode(graph_to_pn(browse_current_graph_nx)),
                        indent=10,
                    )
                )
                st.write("In one line format:")
                st.write(graph_to_pn(browse_current_graph_nx))

            if not st.session_state.df_statistics.empty and st.session_state.sens:
                if st.session_state.sens:
                    nodes, option = rule_chooser()

                st.markdown(
                    f"<span>Result of using all the rules: Precision: <b>{st.session_state.whole_accuracy[0]:.3f}</b>, \
                        Recall: <b>{st.session_state.whole_accuracy[1]:.3f}</b>, Fscore: <b>{st.session_state.whole_accuracy[2]:.3f}</b></span>",
                    unsafe_allow_html=True,
                )

                (
                    fn_graphs,
                    fn_sentences,
                    fn_indices,
                    fp_graphs,
                    fp_sentences,
                    fp_indices,
                    fscore,
                    prec,
                    predicted,
                    recall,
                    support,
                    tp_graphs,
                    tp_sentences,
                    tp_indices,
                ) = extract_data_from_dataframe(option)

                st.markdown(
                    f"<span>The rule's result: Precision: <b>{prec:.3f}</b>, Recall: <b>{recall:.3f}</b>, \
                        Fscore: <b>{fscore:.3f}</b>, True positives: <b>{len(tp_graphs)}</b>, False positives: <b>{len(fp_graphs)}</b></span>",
                    unsafe_allow_html=True,
                )

                tp_fp_fn_choice = (
                    "Predicted",
                    "True Positive graphs",
                    "False Positive graphs",
                    "False Negative graphs",
                )
                tp_fp_fn = st.selectbox(
                    "Select the option you want to view", tp_fp_fn_choice
                )

                if tp_fp_fn == "Predicted":
                    predicted_inds = [
                        i for i, pred in enumerate(predicted) if pred == 1
                    ]
                    filt_df = st.session_state.df[
                        st.session_state.df.index.isin(predicted_inds)
                    ]
                    pred_graphs = filt_df.graph.tolist()
                    pred_sentences = [
                        (sen[0], sen[1])
                        for sen in zip(filt_df.text.tolist(), filt_df.label.tolist())
                    ]
                    graph_viewer(
                        "PD", pred_graphs, pred_sentences, predicted_inds, nodes
                    )

                elif tp_fp_fn == "False Positive graphs":
                    if fp_graphs:
                        graph_viewer("FP", fp_graphs, fp_sentences, fp_indices, nodes)

                elif tp_fp_fn == "True Positive graphs":
                    if tp_graphs:
                        graph_viewer("TP", tp_graphs, tp_sentences, tp_indices, nodes)

                elif tp_fp_fn == "False Negative graphs":
                    if fn_graphs:
                        graph_viewer("FN", fn_graphs, fn_sentences, fn_indices, nodes)


def get_args():
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("-t", "--train-data", type=str)
    parser.add_argument("-v", "--val-data", type=str)
    parser.add_argument(
        "-sr",
        "--suggested-rules",
        default=None,
        type=str,
        help="Rules extracted automatically from python. If not present, the UI will automatically train it.",
    )
    parser.add_argument(
        "-hr",
        "--hand-rules",
        default=None,
        type=str,
        help="Rules extracted with the UI. If provided, the UI will load them.",
    )
    parser.add_argument("-m", "--mode", default="simple", type=str)
    parser.add_argument("-g", "--graph-format", default="fourlang", type=str)
    parser.add_argument(
        "-l",
        "--label",
        default=None,
        type=str,
        help="Specify label for OneVsAll multi-label classification. Datasets require a labels column with all valid labels.",
    )
    return parser.parse_args()


def main(args):
    st.set_page_config(layout="wide")
    hide_streamlit_style = """
            <style>
            #MainMenu {visibility: hidden;}
            footer {
	
                visibility: hidden;
                
                }
            footer:after {
                content:'POTATO'; 
                visibility: visible;
                display: block;
                position: relative;
                #background-color: red;
                padding: 5px;
                top: 2px;
            }
            </style>
            
            """
    st.markdown(hide_streamlit_style, unsafe_allow_html=True)
    st.markdown(
        "<h1 style='text-align: center; color: black;'>POTATO</h1>"
        "<h2 style='text-align: center; color: black;'>explainable information extraction framework</h2>"
        "<h3 style='text-align: center;'>Visit our <a href='https://github.com/adaamko/POTATO'>GitHub</a> page for more information and to access the code</h3><p></p>",
        unsafe_allow_html=True,
    )

    init_session_states()
    evaluator = init_evaluator()
    if args.train_data:
        data = read_df(args.train_data, args.label)
    if args.val_data:
        val_data = read_df(args.val_data, args.label)
    graph_format = args.graph_format
    feature_path = args.suggested_rules
    hand_made_rules = args.hand_rules
    mode = args.mode
    if mode == "simple":
        assert args.train_data, "Train data is required for simple mode"
        assert args.val_data, "Validation data is required for simple mode"
        simple_mode(
            evaluator, data, val_data, graph_format, feature_path, hand_made_rules
        )
    elif mode == "advanced":
        assert args.train_data, "Train data is required for advanced mode"
        advanced_mode(
            evaluator, args.train_data, graph_format, feature_path, hand_made_rules
        )
    # For inference mode you don't need any training data as we will use the model
    # to predict the labels on-the-fly
    elif mode == "inference":
        inference_mode(evaluator, hand_made_rules)

    else:
        st.error("Mode not supported")


if __name__ == "__main__":
    args = get_args()
    main(args)



import re

import numpy as np
from sklearn.tree import _tree


def tree_to_code(tree, feature_graph_strings, inverse_relabel):
    tree_ = tree.tree_
    feature_name = [
        feature_graph_strings[inverse_relabel[int(i)]]
        if i != _tree.TREE_UNDEFINED
        else "undefined!"
        for i in tree_.feature
    ]

    def recurse(node, depth):
        indent = "  " * depth
        if tree_.feature[node] != _tree.TREE_UNDEFINED:
            name = feature_name[node]
            threshold = tree_.threshold[node]
            print("{}if {} <= {}:".format(indent, name, threshold))
            for path in recurse(tree_.children_left[node], depth + 1):
                yield [path[0], path[1] + [name], path[2]]
            print("{}else:  # if {} > {}".format(indent, name, threshold))
            for path in recurse(tree_.children_right[node], depth + 1):
                yield [path[0] + [name], path[1], path[2]]
        else:
            yield [[], [], np.argmax(tree_.value[node])]
            print("{}return {}".format(indent, np.argmax(tree_.value[node])))

    return recurse(0, 1)


def d_clean(string):
    s = string
    for c in "\\=@-,'\".!:;<>/{}[]()#^?":
        s = s.replace(c, "_")
    s = s.replace("$", "_dollars")
    s = s.replace("%", "_percent")
    s = s.replace("|", " ")
    s = s.replace("*", " ")
    if s == "#":
        s = "_number"
    keywords = ("graph", "node", "strict", "edge")
    if re.match("^[0-9]", s) or s in keywords:
        s = "X" + s
    return s


def to_dots(graphs, marked_nodes=set(), integ=False):
    lines = [u"digraph finite_state_machine {", "\tdpi=70;"]
    # lines.append('\tordering=out;')
    # sorting everything to make the process deterministic
    for i, graph in enumerate(graphs):
        s = "subgraph cluster_" + chr(ord("@") + i + 1) + " {"
        node_lines = []

        node_lines.append(s)
        node_to_name = {}
        for node, n_data in graph.nodes(data=True):
            if integ:
                d_node = d_clean(str(node))
            else:
                d_node = d_clean(n_data["name"])
            printname = d_node
            node_to_name[node] = printname
            if (
                "expanded" in n_data
                and n_data["expanded"]
                and printname in marked_nodes
            ):
                node_line = u'\t{0} [shape = circle, label = "{1}", \
                        style=filled, fillcolor=purple];'.format(
                    d_node, printname
                ).replace(
                    "-", "_"
                )
            elif "expanded" in n_data and n_data["expanded"]:
                node_line = u'\t{0} [shape = circle, label = "{1}", \
                        style="filled"];'.format(
                    d_node, printname
                ).replace(
                    "-", "_"
                )
            elif "fourlang" in n_data and n_data["fourlang"]:
                node_line = u'\t{0} [shape = circle, label = "{1}", \
                        style="filled", fillcolor=red];'.format(
                    d_node, printname
                ).replace(
                    "-", "_"
                )
            elif "substituted" in n_data and n_data["substituted"]:
                node_line = u'\t{0} [shape = circle, label = "{1}", \
                        style="filled"];'.format(
                    d_node, printname
                ).replace(
                    "-", "_"
                )
            elif printname in marked_nodes:
                node_line = u'\t{0} [shape = circle, label = "{1}", style=filled, fillcolor=lightblue];'.format(
                    d_node, printname
                ).replace(
                    "-", "_"
                )
            else:
                node_line = u'\t{0} [shape = circle, label = "{1}"];'.format(
                    d_node, printname
                ).replace("-", "_")
            node_lines.append(node_line)
        lines += sorted(node_lines)

        edge_lines = []
        for u, v, edata in graph.edges(data=True):
            if "color" in edata:
                d_node1 = node_to_name[u]
                d_node2 = node_to_name[v]
                edge_lines.append(
                    u'\t{0} -> {1} [ label = "{2}" ];'.format(
                        d_node1, d_node2, edata["color"]
                    )
                )

        lines += sorted(edge_lines)
        lines.append("}")
    lines.append("}")
    return u"\n".join(lines)


def to_dot(graph, marked_nodes=set(), integ=False):
    lines = [u"digraph finite_state_machine {", "\tdpi=70;"]
    # lines.append('\tordering=out;')
    # sorting everything to make the process deterministic
    node_lines = []
    node_to_name = {}
    for node, n_data in graph.nodes(data=True):
        if integ:
            d_node = d_clean(str(node))
        else:
            d_node = d_clean(n_data["name"])
        printname = d_node
        node_to_name[node] = printname
        if "expanded" in n_data and n_data["expanded"] and printname in marked_nodes:
            node_line = u'\t{0} [shape = circle, label = "{1}", \
                    style=filled, fillcolor=purple];'.format(
                d_node, printname
            ).replace(
                "-", "_"
            )
        elif "expanded" in n_data and n_data["expanded"]:
            node_line = u'\t{0} [shape = circle, label = "{1}", \
                    style="filled"];'.format(
                d_node, printname
            ).replace(
                "-", "_"
            )
        elif "fourlang" in n_data and n_data["fourlang"]:
            node_line = u'\t{0} [shape = circle, label = "{1}", \
                    style="filled", fillcolor=red];'.format(
                d_node, printname
            ).replace(
                "-", "_"
            )
        elif "substituted" in n_data and n_data["substituted"]:
            node_line = u'\t{0} [shape = circle, label = "{1}", \
                    style="filled"];'.format(
                d_node, printname
            ).replace(
                "-", "_"
            )
        elif printname in marked_nodes:
            node_line = u'\t{0} [shape = circle, label = "{1}", style=filled, fillcolor=lightblue];'.format(
                d_node, printname
            ).replace(
                "-", "_"
            )
        else:
            node_line = u'\t{0} [shape = circle, label = "{1}"];'.format(
                d_node, printname
            ).replace("-", "_")
        node_lines.append(node_line)
    lines += sorted(node_lines)

    edge_lines = []
    for u, v, edata in graph.edges(data=True):
        if "color" in edata:
            d_node1 = node_to_name[u]
            d_node2 = node_to_name[v]
            edge_lines.append(
                u'\t{0} -> {1} [ label = "{2}" ];'.format(
                    d_node1, d_node2, edata["color"]
                )
            )

    lines += sorted(edge_lines)
    lines.append("}")
    return u"\n".join(lines)



from setuptools import find_packages, setup

setup(
    name="xpotato",
    version="0.1.0",
    description="XAI human-in-the-loop information extraction framework",
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/adaamko/POTATO",
    author="Adam Kovacs, Gabor Recski",
    author_email="adam.kovacs@tuwien.ac.at, gabor.recski@tuwien.ac.at",
    license="MIT",
    install_requires=[
        "pandas >= 1.3.5",
        "tqdm",
        "stanza == 1.3.0",
        "scikit-learn == 1.0.2",
        "eli5 == 0.11.0",
        "graphviz == 0.18.2",
        "penman >= 1.2.1",
        "networkx == 2.6.3",
        "rank_bm25 == 0.2.1",
        "streamlit == 1.3.1",
        "streamlit-aggrid == 0.2.3.post2",
        "scikit-criteria == 0.5",
        "tuw-nlp >= 0.0.4",
        "amrlib == 0.6.0",
    ],
    packages=find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    zip_safe=False,
)



import json
from typing import Dict, List, Union


class Rule:
    def __init__(self, rule, openie=False):
        self.positive_samples = rule[0]
        self.negative_samples = rule[1]
        self.label = rule[2]
        self.openie = openie

        self.marked_nodes = None
        if openie:
            self.marked_nodes = rule[3]

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, Rule):
            return False
        return (
            sorted(self.positive_samples) == sorted(__o.positive_samples)
            and sorted(self.negative_samples) == sorted(__o.negative_samples.sort())
            and self.label == __o.label
        )

    def to_list(self) -> List[List[Union[List[str], str]]]:
        return (
            [self.positive_samples, self.negative_samples, self.label]
            if self.openie == False
            else [
                self.positive_samples,
                self.negative_samples,
                self.label,
                self.marked_nodes,
            ]
        )


class RuleSet:
    def __init__(self, rules: List[Rule] = None):
        if rules is None:
            self.rules = []
        else:
            self.rules = rules

    def __iter__(self):
        for rule in self.rules:
            yield rule

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, RuleSet):
            return False
        return self.rules == __o.rules

    def to_tsv(self, tsv_path: str):
        with open(tsv_path, "w") as f:
            for rule in self.rules:
                positive_samples = ";".join(rule.positive_samples)
                negative_samples = ";".join(rule.negative_samples)
                label = rule.label
                marked_nodes = rule.marked_nodes

                rule_str = f"{label}\t{positive_samples}\t{negative_samples}"

                if rule.openie:
                    rule_str += f"\t{json.dumps(marked_nodes)}"
                f.write(rule_str + "\n")

    def from_tsv(self, tsv_path: str):
        with open(tsv_path, "r") as f:
            for line in f:
                line = line.strip("\n")
                line = line.split("\t")

                positive_samples = [] if line[1] == "" else line[1].split(";")
                negative_samples = [] if line[2] == "" else line[2].split(";")
                label = line[0].strip()
                rule = None
                if len(line) == 3:
                    rule = Rule(
                        [positive_samples, negative_samples, label], openie=False
                    )
                elif len(line) == 4:
                    marked_nodes = [] if line[3] == "" else json.loads(line[3])
                    rule = Rule(
                        [positive_samples, negative_samples, label, marked_nodes],
                        openie=True,
                    )
                else:
                    raise Exception(f"Invalid number of fields: {line}")

                if not rule:
                    raise Exception(f"Invalid rule: {line}")
                self.add_rule(rule)

    def from_dict(
        self, rules: Dict[str, List[List[Union[List[str], str]]]], openie: bool = False
    ):
        for key, value in rules.items():
            for rule in value:
                self.add_rule(Rule(rule, openie=openie))

    def from_json(self, json_path: str, openie: bool = False):
        with open(json_path, "r") as f:
            rules = json.load(f)

        for key, value in rules.items():
            for rule in value:
                self.add_rule(Rule(rule, openie=openie))

    def to_json(self, json_path: str):
        with open(json_path, "w") as f:
            json.dump(self.to_dict(), f)

    def add_rule(self, rule: Rule):
        self.rules.append(rule)

    def to_dict(self) -> Dict[str, List[List[Union[List[str], str]]]]:
        rule_dict = {rule.label: [] for rule in self.rules}

        for rule in self.rules:
            rule_dict[rule.label].append(rule.to_list())

        return rule_dict

    def to_list(self) -> List[List[Union[List[str], str]]]:
        return [rule.to_list() for rule in self.rules]
